@@ -1,8 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2013,2015                                Inria
- * Copyright (C) 2010-2015,2017                           CNRS
- * Copyright (C) 2010,2012-2018                           Université de Bordeaux
+ * Copyright (C) 2013,2015,2017                           Inria
+ * Copyright (C) 2010-2017                                CNRS
+ * Copyright (C) 2010,2012-2019                           Université de Bordeaux
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -18,7 +18,15 @@
 
 #include <starpu.h>
 #include <core/simgrid.h>
+#ifdef STARPU_DEBUG
 #include <core/workers.h>
+#endif
+#include <common/thread.h>
+#include <common/fxt.h>
+#include <common/timing.h>
+
+#include <errno.h>
+#include <limits.h>
 
 #ifdef STARPU_SIMGRID
 #ifdef STARPU_HAVE_XBT_SYNCHRO_H
@@ -50,6 +58,16 @@ static int _starpu_futex_wake = FUTEX_WAKE;
 
 extern int _starpu_simgrid_thread_start(int argc, char *argv[]);
 
+int starpu_pthread_equal(starpu_pthread_t t1, starpu_pthread_t t2)
+{
+	return t1 == t2;
+}
+
+starpu_pthread_t starpu_pthread_self(void)
+{
+	return MSG_process_self();
+}
+
 int starpu_pthread_create_on(char *name, starpu_pthread_t *thread, const starpu_pthread_attr_t *attr STARPU_ATTRIBUTE_UNUSED, void *(*start_routine) (void *), void *arg, msg_host_t host)
 {
 	char **_args;
@@ -241,7 +259,8 @@ int starpu_pthread_setspecific(starpu_pthread_key_t key, const void *pointer)
 	char *end;
 	/* Test whether it is an MPI rank */
 	strtol(process_name, &end, 10);
-	if (!*end)
+	if (!*end || !strcmp(process_name, "wait for mpi transfer") ||
+			(!strcmp(process_name, "main") && _starpu_simgrid_running_smpi()))
 		/* Special-case the SMPI process */
 		array = smpi_process_get_user_data();
 	else
@@ -263,7 +282,8 @@ void* starpu_pthread_getspecific(starpu_pthread_key_t key)
 	char *end;
 	/* Test whether it is an MPI rank */
 	strtol(process_name, &end, 10);
-	if (!*end)
+	if (!*end || !strcmp(process_name, "wait for mpi transfer") ||
+			(!strcmp(process_name, "main") && _starpu_simgrid_running_smpi()))
 		/* Special-case the SMPI processes */
 		array = smpi_process_get_user_data();
 	else
@@ -323,6 +343,30 @@ int starpu_pthread_cond_wait(starpu_pthread_cond_t *cond, starpu_pthread_mutex_t
 	return 0;
 }
 
+int starpu_pthread_cond_timedwait(starpu_pthread_cond_t *cond, starpu_pthread_mutex_t *mutex, const struct timespec *abstime)
+{
+#if SIMGRID_VERSION >= 31800
+	struct timespec now, delta;
+	double delay;
+	int ret = 0;
+	_starpu_clock_gettime(&now);
+	delta.tv_sec = abstime->tv_sec - now.tv_sec;
+	delta.tv_nsec = abstime->tv_nsec - now.tv_nsec;
+	delay = (double) delta.tv_sec + (double) delta.tv_nsec / 1000000000.;
+
+	_STARPU_TRACE_COND_WAIT_BEGIN();
+
+	_starpu_pthread_cond_auto_init(cond);
+	ret = xbt_cond_timedwait(*cond, *mutex, delay) ? ETIMEDOUT : 0;
+
+	_STARPU_TRACE_COND_WAIT_END();
+
+	return ret;
+#else
+	STARPU_ASSERT_MSG(0, "simgrid version is too old for this");
+#endif
+}
+
 int starpu_pthread_cond_destroy(starpu_pthread_cond_t *cond)
 {
 	if (*cond)
@@ -417,12 +461,149 @@ int starpu_pthread_barrier_wait(starpu_pthread_barrier_t *barrier)
 	_STARPU_TRACE_BARRIER_WAIT_END();
 	return 0;
 }
-#endif /* defined(STARPU_SIMGRID) */
+#endif /* defined(STARPU_SIMGRID_HAVE_XBT_BARRIER_INIT) */
+
+int starpu_pthread_queue_init(starpu_pthread_queue_t *q)
+{
+	STARPU_PTHREAD_MUTEX_INIT(&q->mutex, NULL);
+	q->queue = NULL;
+	q->allocqueue = 0;
+	q->nqueue = 0;
+	return 0;
+}
+
+int starpu_pthread_wait_init(starpu_pthread_wait_t *w)
+{
+	STARPU_PTHREAD_MUTEX_INIT(&w->mutex, NULL);
+	STARPU_PTHREAD_COND_INIT(&w->cond, NULL);
+	w->block = 1;
+	return 0;
+}
+
+int starpu_pthread_queue_register(starpu_pthread_wait_t *w, starpu_pthread_queue_t *q)
+{
+	STARPU_PTHREAD_MUTEX_LOCK(&q->mutex);
+
+	if (q->nqueue == q->allocqueue)
+	{
+		/* Make room for the new waiter */
+		unsigned newalloc;
+		newalloc = q->allocqueue * 2;
+		if (!newalloc)
+			newalloc = 1;
+		_STARPU_REALLOC(q->queue, newalloc * sizeof(*(q->queue)));
+		q->allocqueue = newalloc;
+	}
+	q->queue[q->nqueue++] = w;
+
+	STARPU_PTHREAD_MUTEX_UNLOCK(&q->mutex);
+	return 0;
+}
+
+int starpu_pthread_queue_unregister(starpu_pthread_wait_t *w, starpu_pthread_queue_t *q)
+{
+	unsigned i;
+	STARPU_PTHREAD_MUTEX_LOCK(&q->mutex);
+	for (i = 0; i < q->nqueue; i++)
+	{
+		if (q->queue[i] == w)
+		{
+			memmove(&q->queue[i], &q->queue[i+1], (q->nqueue - i - 1) * sizeof(*(q->queue)));
+			break;
+		}
+	}
+	STARPU_ASSERT(i < q->nqueue);
+	q->nqueue--;
+	STARPU_PTHREAD_MUTEX_UNLOCK(&q->mutex);
+	return 0;
+}
+
+int starpu_pthread_wait_reset(starpu_pthread_wait_t *w)
+{
+	STARPU_PTHREAD_MUTEX_LOCK(&w->mutex);
+	w->block = 1;
+	STARPU_PTHREAD_MUTEX_UNLOCK(&w->mutex);
+	return 0;
+}
+
+int starpu_pthread_wait_wait(starpu_pthread_wait_t *w)
+{
+	STARPU_PTHREAD_MUTEX_LOCK(&w->mutex);
+	while (w->block == 1)
+		STARPU_PTHREAD_COND_WAIT(&w->cond, &w->mutex);
+	STARPU_PTHREAD_MUTEX_UNLOCK(&w->mutex);
+	return 0;
+}
+
+/* pthread_cond_timedwait not yet available on windows, but we don't run simgrid there anyway */
+#ifdef STARPU_SIMGRID
+int starpu_pthread_wait_timedwait(starpu_pthread_wait_t *w, const struct timespec *abstime)
+{
+	STARPU_PTHREAD_MUTEX_LOCK(&w->mutex);
+	while (w->block == 1)
+		STARPU_PTHREAD_COND_TIMEDWAIT(&w->cond, &w->mutex, abstime);
+	STARPU_PTHREAD_MUTEX_UNLOCK(&w->mutex);
+	return 0;
+}
+#endif
+
+int starpu_pthread_queue_signal(starpu_pthread_queue_t *q)
+{
+	starpu_pthread_wait_t *w;
+	STARPU_PTHREAD_MUTEX_LOCK(&q->mutex);
+	if (q->nqueue)
+	{
+		/* TODO: better try to wake a sleeping one if possible */
+		w = q->queue[0];
+		STARPU_PTHREAD_MUTEX_LOCK(&w->mutex);
+		w->block = 0;
+		STARPU_PTHREAD_COND_SIGNAL(&w->cond);
+		STARPU_PTHREAD_MUTEX_UNLOCK(&w->mutex);
+	}
+	STARPU_PTHREAD_MUTEX_UNLOCK(&q->mutex);
+	return 0;
+}
+
+int starpu_pthread_queue_broadcast(starpu_pthread_queue_t *q)
+{
+	unsigned i;
+	starpu_pthread_wait_t *w;
+	STARPU_PTHREAD_MUTEX_LOCK(&q->mutex);
+	for (i = 0; i < q->nqueue; i++)
+	{
+		w = q->queue[i];
+		STARPU_PTHREAD_MUTEX_LOCK(&w->mutex);
+		w->block = 0;
+		STARPU_PTHREAD_COND_SIGNAL(&w->cond);
+		STARPU_PTHREAD_MUTEX_UNLOCK(&w->mutex);
+	}
+	STARPU_PTHREAD_MUTEX_UNLOCK(&q->mutex);
+	return 0;
+}
+
+int starpu_pthread_wait_destroy(starpu_pthread_wait_t *w)
+{
+	STARPU_PTHREAD_MUTEX_LOCK(&w->mutex);
+	STARPU_PTHREAD_MUTEX_UNLOCK(&w->mutex);
+	STARPU_PTHREAD_MUTEX_DESTROY(&w->mutex);
+	STARPU_PTHREAD_COND_DESTROY(&w->cond);
+	return 0;
+}
+
+int starpu_pthread_queue_destroy(starpu_pthread_queue_t *q)
+{
+	STARPU_ASSERT(!q->nqueue);
+	STARPU_PTHREAD_MUTEX_LOCK(&q->mutex);
+	STARPU_PTHREAD_MUTEX_UNLOCK(&q->mutex);
+	STARPU_PTHREAD_MUTEX_DESTROY(&q->mutex);
+	free(q->queue);
+	return 0;
+}
 
 #endif /* STARPU_SIMGRID */
 
-#if (defined(STARPU_SIMGRID) && (!defined(STARPU_SIMGRID_HAVE_XBT_BARRIER_INIT) && !defined(xbt_barrier_init))) || (!defined(STARPU_SIMGRID) && !defined(STARPU_HAVE_PTHREAD_BARRIER))
-int starpu_pthread_barrier_init(starpu_pthread_barrier_t *restrict barrier, const starpu_pthread_barrierattr_t *restrict attr, unsigned count)
+#if (defined(STARPU_SIMGRID) && (!defined(STARPU_SIMGRID_HAVE_XBT_BARRIER_INIT)) && !defined(xbt_barrier_init)) || (!defined(STARPU_SIMGRID) && !defined(STARPU_HAVE_PTHREAD_BARRIER))
+int starpu_pthread_barrier_init(starpu_pthread_barrier_t *restrict barrier, const starpu_pthread_barrierattr_t *restrict attr STARPU_ATTRIBUTE_UNUSED, unsigned count)
 {
 	int ret = starpu_pthread_mutex_init(&barrier->mutex, NULL);
 	if (!ret)
@@ -606,34 +787,17 @@ int starpu_pthread_barrier_wait(starpu_pthread_barrier_t *barrier)
  * macros of course) which record when the mutex is held or not */
 int starpu_pthread_mutex_lock_sched(starpu_pthread_mutex_t *mutex)
 {
-	int p_ret = starpu_pthread_mutex_lock(mutex);
-	int workerid = starpu_worker_get_id();
-	if(workerid != -1 && _starpu_worker_mutex_is_sched_mutex(workerid, mutex))
-		_starpu_worker_set_flag_sched_mutex_locked(workerid, 1);
-	return p_ret;
+	return starpu_pthread_mutex_lock(mutex);
 }
 
 int starpu_pthread_mutex_unlock_sched(starpu_pthread_mutex_t *mutex)
 {
-	int workerid = starpu_worker_get_id();
-	if(workerid != -1 && _starpu_worker_mutex_is_sched_mutex(workerid, mutex))
-		_starpu_worker_set_flag_sched_mutex_locked(workerid, 0);
-
 	return starpu_pthread_mutex_unlock(mutex);
 }
 
 int starpu_pthread_mutex_trylock_sched(starpu_pthread_mutex_t *mutex)
 {
-	int ret = starpu_pthread_mutex_trylock(mutex);
-
-	if (!ret)
-	{
-		int workerid = starpu_worker_get_id();
-		if(workerid != -1 && _starpu_worker_mutex_is_sched_mutex(workerid, mutex))
-			_starpu_worker_set_flag_sched_mutex_locked(workerid, 1);
-	}
-
-	return ret;
+	return starpu_pthread_mutex_trylock(mutex);
 }
 
 #ifdef STARPU_DEBUG
@@ -690,6 +854,7 @@ int _starpu_pthread_spin_do_lock(starpu_pthread_spinlock_t *lock)
 	}
 
 	/* We have spent enough time with spinning, let's block */
+	/* This avoids typical 10ms pauses when the application thread tries to submit tasks. */
 	while (1)
 	{
 		/* Tell releaser to wake us */
@@ -759,3 +924,72 @@ void _starpu_pthread_spin_do_unlock(starpu_pthread_spinlock_t *lock)
 #endif
 
 #endif /* defined(STARPU_SIMGRID) || (defined(STARPU_LINUX_SYS) && defined(STARPU_HAVE_XCHG)) || !defined(STARPU_HAVE_PTHREAD_SPIN_LOCK) */
+
+#ifdef STARPU_SIMGRID
+
+int starpu_sem_destroy(starpu_sem_t *sem)
+{
+	MSG_sem_destroy(*sem);
+	return 0;
+}
+
+int starpu_sem_init(starpu_sem_t *sem, int pshared, unsigned value)
+{
+	STARPU_ASSERT_MSG(pshared == 0, "pshared semaphores not supported under simgrid");
+	*sem = MSG_sem_init(value);
+	return 0;
+}
+
+int starpu_sem_post(starpu_sem_t *sem)
+{
+	MSG_sem_release(*sem);
+	return 0;
+}
+
+int starpu_sem_wait(starpu_sem_t *sem)
+{
+	MSG_sem_acquire(*sem);
+	return 0;
+}
+
+int starpu_sem_trywait(starpu_sem_t *sem)
+{
+	if (MSG_sem_would_block(*sem))
+		return EAGAIN;
+	starpu_sem_wait(sem);
+	return 0;
+}
+
+int starpu_sem_getvalue(starpu_sem_t *sem, int *sval)
+{
+#if SIMGRID_VERSION > 31300
+	*sval = MSG_sem_get_capacity(*sem);
+	return 0;
+#else
+	(void) sem;
+	(void) sval;
+	STARPU_ABORT_MSG("sigmrid up to 3.13 did not have working MSG_sem_get_capacity");
+#endif
+}
+
+#elif !defined(_MSC_VER) || defined(BUILDING_STARPU) /* !STARPU_SIMGRID */
+
+int starpu_sem_wait(starpu_sem_t *sem)
+{
+	int ret;
+	while((ret = sem_wait(sem)) == -1 && errno == EINTR)
+		;
+
+	return ret;
+}
+
+int starpu_sem_trywait(starpu_sem_t *sem)
+{
+	int ret;
+	while((ret = sem_trywait(sem)) == -1 && errno == EINTR)
+		;
+	
+	return ret;
+}
+
+#endif
