@@ -1,7 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2010-2017                                Université de Bordeaux
- * Copyright (C) 2010-2013,2015,2017                      CNRS
+ * Copyright (C) 2010-2017,2019                           Université de Bordeaux
+ * Copyright (C) 2010-2013,2015-2017                      CNRS
+ * Copyright (C) 2016                                     Inria
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -26,6 +27,7 @@
 #include <errno.h>
 
 static struct starpu_profiling_worker_info worker_info[STARPU_NMAXWORKERS];
+/* TODO: rather use rwlock */
 static starpu_pthread_mutex_t worker_info_mutex[STARPU_NMAXWORKERS];
 
 /* In case the worker is still sleeping when the user request profiling info,
@@ -49,6 +51,8 @@ struct node_pair
 static int busid_matrix[STARPU_MAXNODES][STARPU_MAXNODES];
 static struct starpu_profiling_bus_info bus_profiling_info[STARPU_MAXNODES][STARPU_MAXNODES];
 static struct node_pair busid_to_node_pair[STARPU_MAXNODES*STARPU_MAXNODES];
+static char bus_direct[STARPU_MAXNODES*STARPU_MAXNODES];
+static int bus_ngpus[STARPU_MAXNODES*STARPU_MAXNODES];
 static unsigned busid_cnt = 0;
 
 static void _starpu_bus_reset_profiling_info(struct starpu_profiling_bus_info *bus_info);
@@ -129,8 +133,6 @@ void _starpu_profiling_init(void)
 	{
 		STARPU_PTHREAD_MUTEX_INIT(&worker_info_mutex[worker], NULL);
 	}
-
-	_starpu_profiling_reset_counters();
 }
 
 void _starpu_profiling_start(void)
@@ -144,7 +146,12 @@ void _starpu_profiling_start(void)
 
 void _starpu_profiling_terminate(void)
 {
+	int worker;
 
+	for (worker = 0; worker < STARPU_NMAXWORKERS; worker++)
+	{
+		STARPU_PTHREAD_MUTEX_DESTROY(&worker_info_mutex[worker]);
+	}
 }
 
 /*
@@ -184,12 +191,13 @@ static void _starpu_worker_reset_profiling_info_with_lock(int workerid)
 	worker_info[workerid].used_cycles = 0;
 	worker_info[workerid].stall_cycles = 0;
 	worker_info[workerid].energy_consumed = 0;
+	worker_info[workerid].flops = 0;
 
 	/* We detect if the worker is already sleeping or doing some
 	 * computation */
 	enum _starpu_worker_status status = _starpu_worker_get_status(workerid);
 
-	if (status == STATUS_SLEEPING || status == STATUS_WAKING_UP)
+	if (status == STATUS_SLEEPING)
 	{
 		worker_registered_sleeping_start[workerid] = 1;
 		_starpu_clock_gettime(&sleeping_start_date[workerid]);
@@ -284,7 +292,8 @@ void _starpu_worker_register_executing_end(int workerid)
 	}
 }
 
-void _starpu_worker_update_profiling_info_executing(int workerid, struct timespec *executing_time, int executed_tasks, uint64_t used_cycles, uint64_t stall_cycles, double energy_consumed)
+
+void _starpu_worker_update_profiling_info_executing(int workerid, struct timespec *executing_time, int executed_tasks, uint64_t used_cycles, uint64_t stall_cycles, double energy_consumed, double flops)
 {
 	if (starpu_profiling_status_get())
 	{
@@ -297,6 +306,7 @@ void _starpu_worker_update_profiling_info_executing(int workerid, struct timespe
 		worker_info[workerid].stall_cycles += stall_cycles;
 		worker_info[workerid].energy_consumed += energy_consumed;
 		worker_info[workerid].executed_tasks += executed_tasks;
+		worker_info[workerid].flops += flops;
 
 		STARPU_PTHREAD_MUTEX_UNLOCK(&worker_info_mutex[workerid]);
 	}
@@ -312,6 +322,7 @@ int starpu_profiling_worker_get_info(int workerid, struct starpu_profiling_worke
 		info->executed_tasks = worker_info[workerid].executed_tasks;
 	}
 
+	STARPU_PTHREAD_MUTEX_LOCK(&_starpu_get_worker_struct(workerid)->sched_mutex);
 	STARPU_PTHREAD_MUTEX_LOCK(&worker_info_mutex[workerid]);
 
 	if (info)
@@ -346,6 +357,7 @@ int starpu_profiling_worker_get_info(int workerid, struct starpu_profiling_worke
 	_starpu_worker_reset_profiling_info_with_lock(workerid);
 
 	STARPU_PTHREAD_MUTEX_UNLOCK(&worker_info_mutex[workerid]);
+	STARPU_PTHREAD_MUTEX_UNLOCK(&_starpu_get_worker_struct(workerid)->sched_mutex);
 
 	return 0;
 }
@@ -398,7 +410,7 @@ static void _starpu_bus_reset_profiling_info(struct starpu_profiling_bus_info *b
 
 int _starpu_register_bus(int src_node, int dst_node)
 {
-	if (busid_matrix[src_node][dst_node] != -1)
+	if (starpu_bus_get_id(src_node, dst_node) != -1)
 		return -EBUSY;
 
 	int busid = STARPU_ATOMIC_ADD(&busid_cnt, 1) - 1;
@@ -434,10 +446,35 @@ int starpu_bus_get_dst(int busid)
 	return busid_to_node_pair[busid].dst;
 }
 
+void starpu_bus_set_direct(int busid, int direct)
+{
+	bus_direct[busid] = direct;
+}
+
+int starpu_bus_get_direct(int busid)
+{
+	return bus_direct[busid];
+}
+
+void starpu_bus_set_ngpus(int busid, int ngpus)
+{
+	bus_ngpus[busid] = ngpus;
+}
+
+int starpu_bus_get_ngpus(int busid)
+{
+	struct _starpu_machine_topology *topology = &_starpu_get_machine_config()->topology;
+	int ngpus = bus_ngpus[busid];
+	if (!ngpus)
+		/* Unknown number of GPUs, assume it's shared by all GPUs */
+		ngpus = topology->ncudagpus+topology->nopenclgpus;
+	return ngpus;
+}
+
 int starpu_bus_get_profiling_info(int busid, struct starpu_profiling_bus_info *bus_info)
 {
-	int src_node = busid_to_node_pair[busid].src;
-	int dst_node = busid_to_node_pair[busid].dst;
+	int src_node = starpu_bus_get_src(busid);
+	int dst_node = starpu_bus_get_dst(busid);
 
 	/* XXX protect all this  method with a mutex */
 	if (bus_info)
