@@ -1,7 +1,7 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2013                                     Inria
- * Copyright (C) 2014-2015,2017                           CNRS
+ * Copyright (C) 2013,2017                                Inria
+ * Copyright (C) 2014-2017                                CNRS
  * Copyright (C) 2014-2018                                Université de Bordeaux
  * Copyright (C) 2013                                     Simon Archipoff
  *
@@ -19,6 +19,7 @@
 
 #include <starpu_sched_component.h>
 #include <starpu_scheduler.h>
+#include <core/workers.h>
 
 #include "fifo_queues.h"
 
@@ -44,16 +45,7 @@ static double fifo_estimated_end(struct starpu_sched_component * component)
 	STARPU_ASSERT(component && component->data);
 	struct _starpu_fifo_data * data = component->data;
 	struct _starpu_fifo_taskq * fifo = data->fifo;
-	starpu_pthread_mutex_t * mutex = &data->mutex;
-	int card = starpu_bitmap_cardinal(component->workers_in_ctx);
-	STARPU_ASSERT(card != 0);
-	double estimated_end = starpu_sched_component_estimated_end_average(component);
-	STARPU_PTHREAD_MUTEX_LOCK(mutex);
-	fifo->exp_start = STARPU_MAX(fifo->exp_start, starpu_timing_now());
-	estimated_end += fifo->exp_start + fifo->exp_len / card;
-	STARPU_PTHREAD_MUTEX_UNLOCK(mutex);
-
-	return estimated_end;
+	return starpu_sched_component_estimated_end_min_add(component, fifo->exp_len);
 }
 
 static double fifo_estimated_load(struct starpu_sched_component * component)
@@ -69,9 +61,9 @@ static double fifo_estimated_load(struct starpu_sched_component * component)
 	{
 		int first_worker = starpu_bitmap_first(component->workers_in_ctx);
 		relative_speedup = starpu_worker_get_relative_speedup(starpu_worker_get_perf_archtype(first_worker, component->tree->sched_ctx_id));
-		STARPU_PTHREAD_MUTEX_LOCK(mutex);
+		STARPU_COMPONENT_MUTEX_LOCK(mutex);
 		load += fifo->ntasks / relative_speedup;
-		STARPU_PTHREAD_MUTEX_UNLOCK(mutex);
+		STARPU_COMPONENT_MUTEX_UNLOCK(mutex);
 		return load;
 	}
 	else
@@ -83,9 +75,9 @@ static double fifo_estimated_load(struct starpu_sched_component * component)
 			relative_speedup += starpu_worker_get_relative_speedup(starpu_worker_get_perf_archtype(i, component->tree->sched_ctx_id));
 		relative_speedup /= starpu_bitmap_cardinal(component->workers_in_ctx);
 		STARPU_ASSERT(!_STARPU_IS_ZERO(relative_speedup));
-		STARPU_PTHREAD_MUTEX_LOCK(mutex);
+		STARPU_COMPONENT_MUTEX_LOCK(mutex);
 		load += fifo->ntasks / relative_speedup;
-		STARPU_PTHREAD_MUTEX_UNLOCK(mutex);
+		STARPU_COMPONENT_MUTEX_UNLOCK(mutex);
 	}
 	return load;
 }
@@ -98,15 +90,14 @@ static int fifo_push_local_task(struct starpu_sched_component * component, struc
 	struct _starpu_fifo_taskq * fifo = data->fifo;
 	starpu_pthread_mutex_t * mutex = &data->mutex;
 	int ret = 0;
-
-	STARPU_PTHREAD_MUTEX_LOCK(mutex);
+	STARPU_COMPONENT_MUTEX_LOCK(mutex);
 	double exp_len;
 	if(!isnan(task->predicted))
 		exp_len = fifo->exp_len + task->predicted;
 	else
 		exp_len = fifo->exp_len;
 
-	if((data->ntasks_threshold != 0) && (data->exp_len_threshold != 0.0) && 
+	if((data->ntasks_threshold != 0) && (data->exp_len_threshold != 0.0) &&
 			((fifo->ntasks >= data->ntasks_threshold) || (exp_len >= data->exp_len_threshold)))
 	{
 		static int warned;
@@ -117,7 +108,7 @@ static int fifo_push_local_task(struct starpu_sched_component * component, struc
 		}
 		STARPU_ASSERT(!is_pushback);
 		ret = 1;
-		STARPU_PTHREAD_MUTEX_UNLOCK(mutex);
+		STARPU_COMPONENT_MUTEX_UNLOCK(mutex);
 	}
 	else
 	{
@@ -137,12 +128,12 @@ static int fifo_push_local_task(struct starpu_sched_component * component, struc
 		STARPU_ASSERT(!isnan(fifo->exp_end));
 		STARPU_ASSERT(!isnan(fifo->exp_len));
 		STARPU_ASSERT(!isnan(fifo->exp_start));
-		
+
+		STARPU_COMPONENT_MUTEX_UNLOCK(mutex);
 		if(!is_pushback)
 			component->can_pull(component);
-		STARPU_PTHREAD_MUTEX_UNLOCK(mutex);
 	}
-	
+
 	return ret;
 }
 
@@ -151,19 +142,20 @@ static int fifo_push_task(struct starpu_sched_component * component, struct star
 	return fifo_push_local_task(component, task, 0);
 }
 
-static struct starpu_task * fifo_pull_task(struct starpu_sched_component * component)
+static struct starpu_task * fifo_pull_task(struct starpu_sched_component * component, struct starpu_sched_component * to STARPU_ATTRIBUTE_UNUSED)
 {
 	STARPU_ASSERT(component && component->data);
 	struct _starpu_fifo_data * data = component->data;
 	struct _starpu_fifo_taskq * fifo = data->fifo;
 	starpu_pthread_mutex_t * mutex = &data->mutex;
-	STARPU_PTHREAD_MUTEX_LOCK(mutex);
+	const double now = starpu_timing_now();
+	STARPU_COMPONENT_MUTEX_LOCK(mutex);
 	struct starpu_task * task = _starpu_fifo_pop_task(fifo, starpu_worker_get_id_check());
 	if(task)
 	{
 		if(!isnan(task->predicted))
 		{
-			fifo->exp_start = starpu_timing_now() + task->predicted;
+			fifo->exp_start = now + task->predicted;
 			fifo->exp_len -= task->predicted;
 		}
 		fifo->exp_end = fifo->exp_start + fifo->exp_len;
@@ -173,23 +165,13 @@ static struct starpu_task * fifo_pull_task(struct starpu_sched_component * compo
 	STARPU_ASSERT(!isnan(fifo->exp_end));
 	STARPU_ASSERT(!isnan(fifo->exp_len));
 	STARPU_ASSERT(!isnan(fifo->exp_start));
-	STARPU_PTHREAD_MUTEX_UNLOCK(mutex);
+	STARPU_COMPONENT_MUTEX_UNLOCK(mutex);
 
 	// When a pop is called, a can_push is called for pushing tasks onto
 	// the empty place of the queue left by the popped task.
-	int i,ret;
-	for(i=0; i < component->nparents; i++)
-	{
-		if(component->parents[i] == NULL)
-			continue;
-		else
-		{
-			ret = component->parents[i]->can_push(component->parents[i]);
-			if(ret)
-				break;
-		}
-	}
-	
+
+	starpu_sched_component_send_can_push_to_parents(component); 
+
 	if(task)
 		return task;
 
@@ -201,29 +183,19 @@ static struct starpu_task * fifo_pull_task(struct starpu_sched_component * compo
  * push fails, which means that the worker fifo_components are
  * currently "full".
  */
-static int fifo_can_push(struct starpu_sched_component * component)
+static int fifo_can_push(struct starpu_sched_component * component, struct starpu_sched_component * to STARPU_ATTRIBUTE_UNUSED)
 {
 	STARPU_ASSERT(component && starpu_sched_component_is_fifo(component));
-	int ret = 0;
 	int res = 0;
+	struct starpu_task * task;
 
-	STARPU_ASSERT(component->nchildren == 1);
-	struct starpu_sched_component * child = component->children[0];
+	task = starpu_sched_component_pump_downstream(component, &res); 
 
-	struct starpu_task * task = fifo_pull_task(component);
 	if(task)
-		ret = starpu_sched_component_push_task(NULL,child,task);	
-	while(task && !ret) 
 	{
-		if(!res)
-			res = 1;
-
-		task = fifo_pull_task(component);
-		if(task)
-			ret = starpu_sched_component_push_task(NULL,child,task);	
+		int ret = fifo_push_local_task(component,task,1);
+		STARPU_ASSERT(!ret);
 	}
-	if(task && ret)
-		fifo_push_local_task(component,task,1); 
 
 	return res;
 }
@@ -235,8 +207,8 @@ int starpu_sched_component_is_fifo(struct starpu_sched_component * component)
 
 struct starpu_sched_component * starpu_sched_component_fifo_create(struct starpu_sched_tree *tree, struct starpu_sched_component_fifo_data * params)
 {
-	struct starpu_sched_component * component = starpu_sched_component_create(tree, "fifo");
-	struct _starpu_fifo_data * data;
+	struct starpu_sched_component *component = starpu_sched_component_create(tree, "fifo");
+	struct _starpu_fifo_data *data;
 	_STARPU_MALLOC(data, sizeof(*data));
 	data->fifo = _starpu_create_fifo();
 	STARPU_PTHREAD_MUTEX_INIT(&data->mutex,NULL);
