@@ -1,8 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2011-2014,2016-2017                      Inria
- * Copyright (C) 2011-2018                                Université de Bordeaux
- * Copyright (C) 2011-2018                                CNRS
+ * Copyright (C) 2011-2014,2016,2017                      Inria
+ * Copyright (C) 2011-2019                                CNRS
+ * Copyright (C) 2011-2019                                Université de Bordeaux
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -101,6 +101,21 @@ int _starpu_codelet_pack_args(void **arg_buffer, size_t *arg_buffer_size, va_lis
 			(void)va_arg(varg_list, void *);
 			(void)va_arg(varg_list, size_t);
 		}
+		else if (arg_type==STARPU_CL_ARGS_NFREE)
+		{
+			(void)va_arg(varg_list, void *);
+			(void)va_arg(varg_list, size_t);
+		}
+		else if (arg_type==STARPU_TASK_DEPS_ARRAY)
+		{
+			(void)va_arg(varg_list, unsigned);
+			(void)va_arg(varg_list, struct starpu_task **);
+		}
+		else if (arg_type==STARPU_TASK_END_DEPS_ARRAY)
+		{
+			(void)va_arg(varg_list, unsigned);
+			(void)va_arg(varg_list, struct starpu_task **);
+		}
 		else if (arg_type==STARPU_CALLBACK)
 		{
 			(void)va_arg(varg_list, _starpu_callback_func_t);
@@ -142,9 +157,13 @@ int _starpu_codelet_pack_args(void **arg_buffer, size_t *arg_buffer_size, va_lis
 		{
 			(void)va_arg(varg_list, starpu_data_handle_t);
 		}
+		else if (arg_type==STARPU_EXECUTE_WHERE)
+		{
+			(void)va_arg(varg_list, unsigned long long);
+		}
 		else if (arg_type==STARPU_EXECUTE_ON_WORKER)
 		{
-			va_arg(varg_list, int);
+			(void)va_arg(varg_list, int);
 		}
 		else if (arg_type==STARPU_WORKER_ORDER)
 		{
@@ -178,6 +197,22 @@ int _starpu_codelet_pack_args(void **arg_buffer, size_t *arg_buffer_size, va_lis
 		{
 			(void)va_arg(varg_list, int);
 		}
+		else if (arg_type==STARPU_TASK_COLOR)
+                {
+                        (void)va_arg(varg_list, int);
+                }
+		else if (arg_type==STARPU_TASK_SYNCHRONOUS)
+                {
+                        (void)va_arg(varg_list, int);
+                }
+		else if (arg_type==STARPU_HANDLES_SEQUENTIAL_CONSISTENCY)
+                {
+                        (void)va_arg(varg_list, unsigned char *);
+                }
+		else if (arg_type==STARPU_TASK_END_DEP)
+		{
+			(void)va_arg(varg_list, int);
+		}
 		else
 		{
 			STARPU_ABORT_MSG("Unrecognized argument %d, did you perhaps forget to end arguments with 0?\n", arg_type);
@@ -216,7 +251,9 @@ void starpu_task_insert_data_make_room(struct starpu_codelet *cl, struct starpu_
 			*allocated_buffers = (current_buffer + room) * 2;
 			_STARPU_REALLOC(task->dyn_handles, *allocated_buffers * sizeof(starpu_data_handle_t));
 			if (cl->nbuffers == STARPU_VARIABLE_NBUFFERS || !cl->dyn_modes)
+			{
 				_STARPU_REALLOC(task->dyn_modes, *allocated_buffers * sizeof(enum starpu_data_access_mode));
+			}
 		}
 	}
 }
@@ -304,6 +341,10 @@ int _starpu_task_insert_create(struct starpu_codelet *cl, struct starpu_task *ta
 	int arg_type;
 	int current_buffer;
 	int allocated_buffers = 0;
+	unsigned ndeps = 0;
+	unsigned nend_deps = 0;
+	struct starpu_task **task_deps_array = NULL;
+	struct starpu_task **task_end_deps_array = NULL;
 
 	_STARPU_TRACE_TASK_BUILD_START();
 
@@ -347,6 +388,24 @@ int _starpu_task_insert_create(struct starpu_codelet *cl, struct starpu_task *ta
 			task->cl_arg_size = va_arg(varg_list, size_t);
 			task->cl_arg_free = 1;
 		}
+		else if (arg_type==STARPU_CL_ARGS_NFREE)
+		{
+			task->cl_arg = va_arg(varg_list, void *);
+			task->cl_arg_size = va_arg(varg_list, size_t);
+			task->cl_arg_free = 0;
+		}
+		else if (arg_type==STARPU_TASK_DEPS_ARRAY)
+		{
+			STARPU_ASSERT_MSG(task_deps_array == NULL, "Parameter 'STARPU_TASK_DEPS_ARRAY' passed twice not supported yet");
+			ndeps = va_arg(varg_list, unsigned);
+			task_deps_array = va_arg(varg_list, struct starpu_task **);
+		}
+		else if (arg_type==STARPU_TASK_END_DEPS_ARRAY)
+		{
+			STARPU_ASSERT_MSG(task_end_deps_array == NULL, "Parameter 'STARPU_TASK_END_DEPS_ARRAY' passed twice not supported yet");
+			nend_deps = va_arg(varg_list, unsigned);
+			task_end_deps_array = va_arg(varg_list, struct starpu_task **);
+		}
 		else if (arg_type==STARPU_CALLBACK)
 		{
 			task->callback_func = va_arg(varg_list, _starpu_callback_func_t);
@@ -390,6 +449,10 @@ int _starpu_task_insert_create(struct starpu_codelet *cl, struct starpu_task *ta
 		{
 			(void)va_arg(varg_list, starpu_data_handle_t);
 		}
+		else if (arg_type==STARPU_EXECUTE_WHERE)
+		{
+			task->where = va_arg(varg_list, unsigned long long);
+		}
 		else if (arg_type==STARPU_EXECUTE_ON_WORKER)
 		{
 			int worker = va_arg(varg_list, int);
@@ -448,6 +511,23 @@ int _starpu_task_insert_create(struct starpu_codelet *cl, struct starpu_task *ta
 		{
 			(void)va_arg(varg_list, int);
 		}
+		else if (arg_type==STARPU_TASK_COLOR)
+                {
+                        task->color = va_arg(varg_list, int);
+                }
+		else if (arg_type==STARPU_TASK_SYNCHRONOUS)
+                {
+                        task->synchronous = va_arg(varg_list, int);
+                }
+		else if (arg_type==STARPU_HANDLES_SEQUENTIAL_CONSISTENCY)
+                {
+                        task->handles_sequential_consistency = va_arg(varg_list, unsigned char *);
+                }
+		else if (arg_type==STARPU_TASK_END_DEP)
+		{
+			int end_dep = va_arg(varg_list, int);
+			starpu_task_end_dep_add(task, end_dep);
+		}
 		else
 		{
 			STARPU_ABORT_MSG("Unrecognized argument %d, did you perhaps forget to end arguments with 0?\n", arg_type);
@@ -477,6 +557,16 @@ int _starpu_task_insert_create(struct starpu_codelet *cl, struct starpu_task *ta
 		starpu_codelet_pack_arg_fini(&state, &task->cl_arg, &task->cl_arg_size);
 	}
 
+	if (task_deps_array)
+	{
+		starpu_task_declare_deps_array(task, ndeps, task_deps_array);
+	}
+
+	if (task_end_deps_array)
+	{
+		starpu_task_declare_end_deps_array(task, nend_deps, task_end_deps_array);
+	}
+
 	_STARPU_TRACE_TASK_BUILD_END();
 	return 0;
 }
@@ -486,6 +576,10 @@ int _fstarpu_task_insert_create(struct starpu_codelet *cl, struct starpu_task *t
 	int arg_i = 0;
 	int current_buffer = 0;
 	int allocated_buffers = 0;
+	unsigned ndeps = 0;
+	unsigned nend_deps = 0;
+	struct starpu_task **task_deps_array = NULL;
+	struct starpu_task **task_end_deps_array = NULL;
 
 	_STARPU_TRACE_TASK_BUILD_START();
 
@@ -539,6 +633,30 @@ int _fstarpu_task_insert_create(struct starpu_codelet *cl, struct starpu_task *t
 			task->cl_arg_size = (size_t)(intptr_t)arglist[arg_i];
 			task->cl_arg_free = 1;
 		}
+		else if (arg_type == STARPU_CL_ARGS_NFREE)
+		{
+			arg_i++;
+			task->cl_arg = arglist[arg_i];
+			arg_i++;
+			task->cl_arg_size = (size_t)(intptr_t)arglist[arg_i];
+			task->cl_arg_free = 0;
+		}
+		else if (arg_type==STARPU_TASK_DEPS_ARRAY)
+		{
+			STARPU_ASSERT_MSG(task_deps_array == NULL, "Parameter 'STARPU_TASK_DEPS_ARRAY' passed twice not supported yet");
+			arg_i++;
+			ndeps = *(unsigned *)arglist[arg_i];
+			arg_i++;
+			task_deps_array = arglist[arg_i];
+		}
+		else if (arg_type==STARPU_TASK_END_DEPS_ARRAY)
+		{
+			STARPU_ASSERT_MSG(task_end_deps_array == NULL, "Parameter 'STARPU_TASK_END_DEPS_ARRAY' passed twice not supported yet");
+			arg_i++;
+			nend_deps = *(unsigned *)arglist[arg_i];
+			arg_i++;
+			task_end_deps_array = arglist[arg_i];
+		}
 		else if (arg_type == STARPU_CALLBACK)
 		{
 			arg_i++;
@@ -591,6 +709,13 @@ int _fstarpu_task_insert_create(struct starpu_codelet *cl, struct starpu_task *t
 			arg_i++;
 			(void)arglist[arg_i];
 		}
+		else if (arg_type == STARPU_EXECUTE_WHERE)
+		{
+			assert(0);
+			arg_i++;
+			unsigned long long where = *(unsigned long long *)arglist[arg_i];
+			task->where = where;
+		}
 		else if (arg_type == STARPU_EXECUTE_ON_WORKER)
 		{
 			arg_i++;
@@ -652,6 +777,25 @@ int _fstarpu_task_insert_create(struct starpu_codelet *cl, struct starpu_task *t
 			arg_i++;
 			(void)arglist[arg_i];
 		}
+		else if (arg_type == STARPU_TASK_COLOR)
+                {
+                        arg_i++;
+                        task->color = *(int *)arglist[arg_i];
+                }
+		else if (arg_type == STARPU_TASK_SYNCHRONOUS)
+                {
+                        arg_i++;
+                        task->synchronous = *(int *)arglist[arg_i];
+                }
+		else if (arg_type==STARPU_HANDLES_SEQUENTIAL_CONSISTENCY)
+                {
+                        task->handles_sequential_consistency = (unsigned char *)arglist[arg_i];
+                }
+		else if (arg_type==STARPU_TASK_END_DEP)
+		{
+			arg_i++;
+			starpu_task_end_dep_add(task, *(int*)arglist[arg_i]);
+		}
 		else
 		{
 			STARPU_ABORT_MSG("unknown/unsupported argument %d, did you perhaps forget to end arguments with 0?", arg_type);
@@ -682,6 +826,16 @@ int _fstarpu_task_insert_create(struct starpu_codelet *cl, struct starpu_task *t
 		starpu_codelet_pack_arg_fini(&state, &task->cl_arg, &task->cl_arg_size);
 	}
 
+	if (task_deps_array)
+	{
+		starpu_task_declare_deps_array(task, ndeps, task_deps_array);
+	}
+
+	if (task_end_deps_array)
+	{
+		starpu_task_declare_end_deps_array(task, nend_deps, task_end_deps_array);
+	}
+
 	_STARPU_TRACE_TASK_BUILD_END();
 
 	return 0;
