@@ -1,7 +1,7 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2012-2013                                Inria
  * Copyright (C) 2010-2015,2017                           CNRS
+ * Copyright (C) 2012,2013,2017                           Inria
  * Copyright (C) 2009-2012,2014,2017                      Université de Bordeaux
  *
  * StarPU is free software; you can redistribute it and/or modify
@@ -452,6 +452,8 @@ STARPUFFT(showstats)(FILE *out)
 	for (worker = 0, total = 0; worker < starpu_worker_get_count(); worker++)
 		total += task_per_worker[worker];
 
+	if (!total)
+		return;
 	for (worker = 0; worker < starpu_worker_get_count(); worker++)
 	{
 		if (task_per_worker[worker])
@@ -503,6 +505,20 @@ STARPUFFT(report_error)(const char *func, const char *file, int line, cufftResul
 	case CUFFT_UNALIGNED_DATA:
 		errormsg = "unaligned data";
 		break;
+#if defined(MAX_CUFFT_ERROR) && (MAX_CUFFT_ERROR >= 0xE)
+	case CUFFT_INCOMPLETE_PARAMETER_LIST:
+		errormsg = "incomplete parameter list";
+		break;
+	case CUFFT_INVALID_DEVICE:
+		errormsg = "invalid device";
+		break;
+	case CUFFT_PARSE_ERROR:
+		errormsg = "parse error";
+		break;
+	case CUFFT_NO_WORKSPACE:
+		errormsg = "no workspace";
+		break;
+#endif /* MAX_CUFFT_ERROR >= 0xE */
 	default:
 		errormsg = "unknown error";
 		break;
