@@ -1,6 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2012 INRIA
+ * Copyright (C) 2012                                     Inria
+ * Copyright (C) 2012,2013,2015,2017                      CNRS
+ * Copyright (C) 2014-2016                                Université de Bordeaux
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -47,6 +49,8 @@ run(struct starpu_sched_policy *policy)
 	{
 		.cpu_funcs = {dummy},
 		.cpu_funcs_name = {"dummy"},
+		.opencl_funcs = {dummy},
+		.cuda_funcs = {dummy},
 		.nbuffers = 0
 	};
 
@@ -84,8 +88,7 @@ enodev:
 	return -ENODEV;
 }
 
-int
-main(void)
+int main(void)
 {
 	struct starpu_sched_policy **policies;
 	struct starpu_sched_policy **policy;
